{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# BigBench-Hard\n",
    "\n",
    "## Introduction\n",
    "\n",
    "In this notebook, we will demonstrate how to use the `trace` package to optimize prompts and code for natural language processing tasks using the BigBench-Hard benchmark. SotA approaches on this benchmark only optimize prompts, while relying on hand-written code to extract answers from LLM responses. By leveraging the LLM-based optimizers provided in `trace`, we aim to enhance the performance of a workflow calling LLMs and post-processing their responses in generating accurate and relevant answers.\n",
    "\n",
    "## Setup\n",
    "\n",
    "First, we'll import the necessary packages and set up our environment. We will use a copy of the BigBench-Hard benchmark hosted on [HuggingFace](https://huggingface.co/datasets/maveriq/bigbenchhard). To use HuggingFace datasets, ensure that you have the `datasets` package installed:\n",
    "\n",
    "pip install datasets"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "outputs": [],
   "source": [
    "!pip install trace-opt"
   ],
   "metadata": {
    "collapsed": false
   }
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:03.229950Z",
     "start_time": "2024-07-19T21:18:59.801910Z"
    }
   },
   "outputs": [],
   "source": [
    "# Import necessary libraries\n",
    "import autogen\n",
    "from opto.trace.nodes import node, GRAPH, ParameterNode\n",
    "from opto.optimizers import OptoPrime\n",
    "from datasets import load_dataset\n",
    "from textwrap import dedent\n",
    "from opto.trace.bundle import bundle\n",
    "from opto.trace.modules import model\n",
    "from opto.trace.errors import ExecutionError\n",
    "from opto.trace.nodes import ExceptionNode\n",
    "from typing import List\n",
    "import re"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Define the Evaluation Function\n",
    "\n",
    "Next, we'll define the utility function for evaluating answers obtained by prompting an LLM."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:03.230792Z",
     "start_time": "2024-07-19T21:19:03.223927Z"
    }
   },
   "outputs": [],
   "source": [
    "def eval_metric(true, prediction):\n",
    "    matches = re.findall(r\"\\([A-Z]\\)\", true)\n",
    "    if matches:\n",
    "        pred = prediction\n",
    "        matches = re.findall(r\"\\([A-Z]\\)\", pred)\n",
    "        parsed_answer = matches[-1] if matches else \"\"\n",
    "        return parsed_answer == true\n",
    "    else:\n",
    "        return prediction == true"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Helper Function\n",
    "\n",
    "We'll create a helper class called `LLMCallable` to interact with the LLM API."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:03.880915Z",
     "start_time": "2024-07-19T21:19:03.657068Z"
    }
   },
   "outputs": [],
   "source": [
    "class LLMCallable:\n",
    "    def __init__(self, config_list=None, max_tokens=1024, verbose=False):\n",
    "        if config_list is None:\n",
    "            config_list = autogen.config_list_from_json(\"OAI_CONFIG_LIST\")\n",
    "        self.llm = autogen.OpenAIWrapper(config_list=config_list)\n",
    "        self.max_tokens = max_tokens\n",
    "        self.verbose = verbose\n",
    "\n",
    "    @bundle(catch_execution_error=True)\n",
    "    def call_llm(self, user_prompt):\n",
    "        system_prompt = \"You are a helpful assistant.\\n\"\n",
    "        messages = [{\"role\": \"system\", \"content\": system_prompt}, {\"role\": \"user\", \"content\": user_prompt}]\n",
    "        response = self.llm.create(messages=messages, max_tokens=self.max_tokens)\n",
    "        response = response.choices[0].message.content\n",
    "\n",
    "        if self.verbose:\n",
    "            print(\"LLM response:\\n\", response)\n",
    "        return response\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Define a Traced Class\n",
    "\n",
    "We will define a Predict class to generate predictions using LLM. Note that we use a module provided by `trace` called `Model` which can wrap a python class to enable tracing. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:04.970205Z",
     "start_time": "2024-07-19T21:19:04.933904Z"
    }
   },
   "outputs": [],
   "source": [
    "@model\n",
    "class Predict(LLMCallable):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "\n",
    "        self.demos = []\n",
    "        self. prompt_template = dedent(\n",
    "            \"\"\"\n",
    "        Given the fields `question`, produce the fields `answer`.\n",
    "\n",
    "        ---\n",
    "\n",
    "        Follow the following format.\n",
    "\n",
    "        Question: \n",
    "        Answer: \n",
    "\n",
    "        ---\n",
    "        Question: {}\n",
    "        Answer:\n",
    "        \"\"\"\n",
    "        )\n",
    "        self.prompt_template = ParameterNode(self.prompt_template, trainable=True,\n",
    "                                             description=\"This is the Prompt Template to the LLM. \" + \\\n",
    "                                                         \"Need to include information about what the format of answers LLM should output. \" + \\\n",
    "                                                         \"They can be (A)/(B), a number like 8, or a string, or Yes/No.\")\n",
    "\n",
    "    @bundle(trainable=True, catch_execution_error=True, allow_external_dependencies=True)\n",
    "    def extract_answer(self, prompt_template, question, response):\n",
    "        answer = response.split(\"Answer:\")[1].strip()\n",
    "        return answer\n",
    "\n",
    "    @bundle(trainable=True, catch_execution_error=True, allow_external_dependencies=True)\n",
    "    def create_prompt(self, prompt_template, question):\n",
    "        return prompt_template.format(question)\n",
    "\n",
    "    def forward(self, question):\n",
    "        user_prompt = self.create_prompt(self.prompt_template, question)\n",
    "        response = self.call_llm(user_prompt)\n",
    "        answer = self.extract_answer(self.prompt_template, question, response)\n",
    "        return answer"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Define the optimizer\n",
    "\n",
    "Note that the `prompt_template` is a `ParameterNode` as well as the `extract_answer` is a trainable function. `trace` handles the optimization of heterogenous parameters seamlessly."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:06.148198Z",
     "start_time": "2024-07-19T21:19:06.128481Z"
    }
   },
   "outputs": [],
   "source": [
    "def learn_predict(dp, optimizer, examples):\n",
    "    for step, example in enumerate(examples):\n",
    "        try:\n",
    "            response = dp.forward(example['question'])\n",
    "            correctness = eval_metric(example['answer'], response)\n",
    "            feedback = \"The answer is correct! No need to change anything.\" if correctness else f\"The answer is wrong. We expect the output of your answer to be \\\"{example['answer']}\\\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer.\"\n",
    "        except ExecutionError as e:\n",
    "            response = e.exception_node\n",
    "            feedback = response.data\n",
    "            correctness = False\n",
    "            \n",
    "        print(\"Question:\", example[\"question\"])\n",
    "        print(\"Expected answer:\", example[\"answer\"])\n",
    "        print(\"Answer:\", response)\n",
    "\n",
    "        if correctness:\n",
    "            continue\n",
    "\n",
    "        optimizer.zero_feedback()\n",
    "        optimizer.backward(response, feedback)\n",
    "\n",
    "        print(f\"Output: {response}, Feedback: {feedback}, Variables:\")  # Logging\n",
    "        for p in optimizer.parameters:\n",
    "            print(p.name, p.data)\n",
    "        optimizer.step(verbose=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Putting it all together\n",
    "\n",
    "Finally, we use the optimizer to find better prompts using a small training set as follows."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "ExecuteTime": {
     "end_time": "2024-07-19T21:19:21.867979Z",
     "start_time": "2024-07-19T21:19:07.382684Z"
    }
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Training on a few examples:\n",
      "Question: Is the following sentence plausible? \"Elias Lindholm beat the buzzer.\"\n",
      "Expected answer: no\n",
      "Answer: MessageNode: (eval:1, dtype=<class 'str'>, data=Yes)\n",
      "Output: MessageNode: (eval:1, dtype=<class 'str'>, data=Yes), Feedback: The answer is wrong. We expect the output of your answer to be \"no\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer., Variables:\n",
      "__code:1 def create_prompt(self, prompt_template, question):\n",
      "        return prompt_template.format(question)\n",
      "__code:0 def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip()\n",
      "        return answer\n",
      "str:0 \n",
      "Given the fields `question`, produce the fields `answer`.\n",
      "\n",
      "---\n",
      "\n",
      "Follow the following format.\n",
      "\n",
      "Question: \n",
      "Answer: \n",
      "\n",
      "---\n",
      "Question: {}\n",
      "Answer:\n",
      "\n",
      "str:0 \n",
      "Given the fields `question`, produce the fields `answer`.\n",
      "\n",
      "---\n",
      "\n",
      "Follow the following format.\n",
      "\n",
      "Question: \n",
      "Answer: \n",
      "\n",
      "---\n",
      "Question: {}\n",
      "Answer:\n",
      "\n",
      "Prompt\n",
      " \n",
      "You're tasked to solve a coding/algorithm problem. You will see the instruction, the code, the documentation of each function used in the code, and the feedback about the execution result.\n",
      "\n",
      "Specifically, a problem will be composed of the following parts:\n",
      "- #Instruction: the instruction which describes the things you need to do or the question you should answer.\n",
      "- #Code: the code defined in the problem.\n",
      "- #Documentation: the documentation of each function used in #Code. The explanation might be incomplete and just contain high-level description. You can use the values in #Others to help infer how those functions work.\n",
      "- #Variables: the input variables that you can change.\n",
      "- #Constraints: the constraints or descriptions of the variables in #Variables.\n",
      "- #Inputs: the values of other inputs to the code, which are not changeable.\n",
      "- #Others: the intermediate values created through the code execution.\n",
      "- #Outputs: the result of the code output.\n",
      "- #Feedback: the feedback about the code's execution result.\n",
      "\n",
      "In #Variables, #Inputs, #Outputs, and #Others, the format is:\n",
      "\n",
      "<data_type> <variable_name> = <value>\n",
      "\n",
      "If <type> is (code), it means <value> is the source code of a python code, which may include docstring and definitions.\n",
      "\n",
      "Output_format: Your output should be in the following json format, satisfying the json syntax:\n",
      "\n",
      "{{\n",
      "\"reasoning\": <Your reasoning>,\n",
      "\"answer\": <Your answer>,\n",
      "\"suggestion\": {{\n",
      "    <variable_1>: <suggested_value_1>,\n",
      "    <variable_2>: <suggested_value_2>,\n",
      "}}\n",
      "}}\n",
      "\n",
      "In \"reasoning\", explain the problem: 1. what the #Instruction means 2. what the #Feedback on #Output means to #Variables considering how #Variables are used in #Code and other values in #Documentation, #Inputs, #Others. 3. Reasoning about the suggested changes in #Variables (if needed) and the expected result.\n",
      "\n",
      "If #Instruction asks for an answer, write it down in \"answer\".\n",
      "\n",
      "If you need to suggest a change in the values of #Variables, write down the suggested values in \"suggestion\". Remember you can change only the values in #Variables, not others. When <type> of a variable is (code), you should write the new definition in the format of python code without syntax errors, and you should not change the function name or the function signature.\n",
      "\n",
      "If no changes or answer are needed, just output TERMINATE.\n",
      "\n",
      "Now you see problem instance:\n",
      "\n",
      "================================\n",
      "\n",
      "#Instruction\n",
      "You need to change the <value> of the variables in #Variables to improve the output in accordance to #Feedback.\n",
      "\n",
      "#Code\n",
      "eval0 = eval(self=ModelWrapper0, prompt_template=str0, question=str0, __code=__code1)\n",
      "LLMCallable.call_llm0 = LLMCallable.call_llm(self=ModelWrapper1, user_prompt=eval0)\n",
      "eval1 = eval(self=ModelWrapper2, prompt_template=str0, question=str1, response=LLMCallable.call_llm0, __code=__code0)\n",
      "\n",
      "#Documentation\n",
      "[eval] This operator eval(__code, *args, **kwargs) evaluates the code block, where __code is the code (str) and *args and **kwargs are the arguments of the function. The output is the result of the evaluation, i.e., __code(*args, **kwargs).\n",
      "[LLMCallable.call_llm] .\n",
      "\n",
      "#Variables\n",
      "(str) str0=Is the following sentence plausible? \"Elias Lindholm beat the buzzer.\"\n",
      "(code) __code1:def create_prompt(self, prompt_template, question):\n",
      "        return prompt_template.format(question)\n",
      "(code) __code0:def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip()\n",
      "        return answer\n",
      "\n",
      "#Constraints\n",
      "(code) __code1: The code should start with:\n",
      "def create_prompt(self, prompt_template, question):\n",
      "(code) __code0: The code should start with:\n",
      "def extract_answer(self, prompt_template, question, response):\n",
      "\n",
      "#Inputs\n",
      "(ModelWrapper) ModelWrapper2=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(str) str1=Is the following sentence plausible? \"Elias Lindholm beat the buzzer.\"\n",
      "(ModelWrapper) ModelWrapper1=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(ModelWrapper) ModelWrapper0=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "\n",
      "#Others\n",
      "(str) eval0=\n",
      "Given the fields `question`, produce the fields `answer`.\n",
      "\n",
      "---\n",
      "\n",
      "Follow the following format.\n",
      "\n",
      "Question: \n",
      "Answer: \n",
      "\n",
      "---\n",
      "Question: Is the following sentence plausible? \"Elias Lindholm beat the buzzer.\"\n",
      "Answer:\n",
      "\n",
      "(str) LLMCallable.call_llm0=Question: Is the following sentence plausible? \"Elias Lindholm beat the buzzer.\"\n",
      "Answer: Yes\n",
      "\n",
      "#Outputs\n",
      "(str) eval1=Yes\n",
      "\n",
      "#Feedback:\n",
      "The answer is wrong. We expect the output of your answer to be \"no\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer.\n",
      "\n",
      "================================\n",
      "\n",
      "\n",
      "Your response:\n",
      "\n",
      "LLM response:\n",
      " {\n",
      "\"reasoning\": \"The feedback indicates that the current settings and prompt formulation lead to an incorrect answer of 'Yes' from the language model (LLM) regarding the question's plausibility. The question was whether it's plausible for 'Elias Lindholm beat the buzzer.' The output we need from this model is 'no'. To address this, we need to adjust either how the prompt is being constructed (__code1) or how the response is being handled (__code0). However, the issue seems more aligned with how we form the question to the model rather than how we extract the answer, since the model outputs 'Yes', which means it confirms plausibility based on the question's current phrasing or context. Since the issue is about generating the right context for the LLM to produce 'no', the suggestion is to modify the prompt template or the way the prompt is constructed to provide more contextual clues to the LLM that could push it towards a 'no' answer. There were no negative remarks on the code's ability to call the LLM or extract the answer, suggesting the logic for these operations is correct. Thus, we'll adjust the question's phrasing or add to the template to help give the model context that leads to realizing the scenario in the question is implausible.\",\n",
      "\"answer\": \"\",\n",
      "\"suggestion\": {\n",
      "\"str0\": \"Considering known historical contexts and logical sequences, is the following sentence plausible? \\\"Elias Lindhorn beat the buzzer.\\\"\",\n",
      "\"__code1\": \"def create_prompt(self, prompt_template, question):\\n       return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question}'\"\n",
      "}\n",
      "}\n",
      "Question: Is the following sentence plausible? \"John Carlson scored in the third period.\"\n",
      "Expected answer: yes\n",
      "Answer: MessageNode: (exception_eval:0, dtype=<class 'str'>, data=def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip() <--- (IndexError) list index out of range\n",
      "        return answer)\n",
      "Output: MessageNode: (exception_eval:0, dtype=<class 'str'>, data=def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip() <--- (IndexError) list index out of range\n",
      "        return answer), Feedback: def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip() <--- (IndexError) list index out of range\n",
      "        return answer, Variables:\n",
      "__code:1 def create_prompt(self, prompt_template, question):\n",
      "       return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question}'\n",
      "__code:0 def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip()\n",
      "        return answer\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "Prompt\n",
      " \n",
      "You're tasked to solve a coding/algorithm problem. You will see the instruction, the code, the documentation of each function used in the code, and the feedback about the execution result.\n",
      "\n",
      "Specifically, a problem will be composed of the following parts:\n",
      "- #Instruction: the instruction which describes the things you need to do or the question you should answer.\n",
      "- #Code: the code defined in the problem.\n",
      "- #Documentation: the documentation of each function used in #Code. The explanation might be incomplete and just contain high-level description. You can use the values in #Others to help infer how those functions work.\n",
      "- #Variables: the input variables that you can change.\n",
      "- #Constraints: the constraints or descriptions of the variables in #Variables.\n",
      "- #Inputs: the values of other inputs to the code, which are not changeable.\n",
      "- #Others: the intermediate values created through the code execution.\n",
      "- #Outputs: the result of the code output.\n",
      "- #Feedback: the feedback about the code's execution result.\n",
      "\n",
      "In #Variables, #Inputs, #Outputs, and #Others, the format is:\n",
      "\n",
      "<data_type> <variable_name> = <value>\n",
      "\n",
      "If <type> is (code), it means <value> is the source code of a python code, which may include docstring and definitions.\n",
      "\n",
      "Output_format: Your output should be in the following json format, satisfying the json syntax:\n",
      "\n",
      "{{\n",
      "\"reasoning\": <Your reasoning>,\n",
      "\"answer\": <Your answer>,\n",
      "\"suggestion\": {{\n",
      "    <variable_1>: <suggested_value_1>,\n",
      "    <variable_2>: <suggested_value_2>,\n",
      "}}\n",
      "}}\n",
      "\n",
      "In \"reasoning\", explain the problem: 1. what the #Instruction means 2. what the #Feedback on #Output means to #Variables considering how #Variables are used in #Code and other values in #Documentation, #Inputs, #Others. 3. Reasoning about the suggested changes in #Variables (if needed) and the expected result.\n",
      "\n",
      "If #Instruction asks for an answer, write it down in \"answer\".\n",
      "\n",
      "If you need to suggest a change in the values of #Variables, write down the suggested values in \"suggestion\". Remember you can change only the values in #Variables, not others. When <type> of a variable is (code), you should write the new definition in the format of python code without syntax errors, and you should not change the function name or the function signature.\n",
      "\n",
      "If no changes or answer are needed, just output TERMINATE.\n",
      "\n",
      "Now you see problem instance:\n",
      "\n",
      "================================\n",
      "\n",
      "#Instruction\n",
      "You need to change the <value> of the variables in #Variables to improve the output in accordance to #Feedback.\n",
      "\n",
      "#Code\n",
      "eval0 = eval(self=ModelWrapper0, prompt_template=str0, question=str0, __code=__code1)\n",
      "LLMCallable.call_llm0 = LLMCallable.call_llm(self=ModelWrapper1, user_prompt=eval0)\n",
      "exception_eval0 = eval(self=ModelWrapper2, prompt_template=str0, question=str1, response=LLMCallable.call_llm0, __code=__code0)\n",
      "\n",
      "#Documentation\n",
      "[exception] The operator eval raises an exception.\n",
      "[LLMCallable.call_llm] .\n",
      "\n",
      "#Variables\n",
      "(str) str0=Is the following sentence plausible? \"John Carlson scored in the third period.\"\n",
      "(code) __code1:def create_prompt(self, prompt_template, question):\n",
      "       return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question}'\n",
      "(code) __code0:def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip()\n",
      "        return answer\n",
      "\n",
      "#Constraints\n",
      "(code) __code1: The code should start with:\n",
      "def create_prompt(self, prompt_template, question):\n",
      "(code) __code0: The code should start with:\n",
      "def extract_answer(self, prompt_template, question, response):\n",
      "\n",
      "#Inputs\n",
      "(ModelWrapper) ModelWrapper2=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(str) str1=Is the following sentence plausible? \"John Carlson scored in the third period.\"\n",
      "(ModelWrapper) ModelWrapper1=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(ModelWrapper) ModelWrapper0=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "\n",
      "#Others\n",
      "(str) eval0=Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\" With a focus on the likely scenarios, evaluate: Is the following sentence plausible? \"John Carlson scored in the third period.\"\n",
      "(str) LLMCallable.call_llm0=To evaluate the plausibility of both sentences, let's break them down considering the contexts in which they might occur.\n",
      "\n",
      "1. **\"Elias Lindhorn beat the buzzer.\"**\n",
      "   - Plausibility: Yes, this sentence is plausible.\n",
      "   - Context: This sentence likely refers to a situation in a basketball game or any sport where a buzzer marks the end of a period, quarter, or game. \"Beating the buzzer\" means scoring just before the game or period ends. It's a common phrase in sports commentary. Assuming Elias Lindhorn is a fictional or actual athlete in a sport that features a buzzer, this scenario is very plausible.\n",
      "\n",
      "2. **\"John Carlson scored in the third period.\"**\n",
      "   - Plausibility: Yes, this sentence is also plausible.\n",
      "   - Context: This sentence is likely referring to a hockey game since hockey matches are divided into three periods and John Carlson is a name that could be associated with a hockey player. Scoring in the third period is a straightforward event that happens regularly in hockey games. Given that \"John Carlson\" could also refer to a real-life NHL defenseman known for his offensive skills, this makes the sentence not just plausible but likely in the context of a hockey match.\n",
      "\n",
      "In summary, both sentences are plausible within their respective sports contexts (basketball or other buzzer sports for the first, and hockey for the second), and they logically fit into the narrative flows of those sports.\n",
      "\n",
      "#Outputs\n",
      "(str) exception_eval0=def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip() <--- (IndexError) list index out of range\n",
      "        return answer\n",
      "\n",
      "#Feedback:\n",
      "def extract_answer(self, prompt_template, question, response):\n",
      "        answer = response.split(\"Answer:\")[1].strip() <--- (IndexError) list index out of range\n",
      "        return answer\n",
      "\n",
      "================================\n",
      "\n",
      "\n",
      "Your response:\n",
      "\n",
      "LLM response:\n",
      " {\n",
      "\"reasoning\": \"The function extract_answer is designed to split the response by 'Answer:' and then select the part after this split to extract the answer. However, the IndexError exception indicates that the response string does not contain the substring 'Answer:', leading to an attempt to access a list index that does not exist, since splitting by 'Answer:' does not generate a second element in the list. To correct this, we either need to ensure the response includes the term 'Answer:' followed by the actual answer or modify the extract_answer function to handle cases when 'Answer:' is not present. Considering that changing the response output of LLMCallable.call_llm would be external to this specific task and could affect other dependencies, adjusting the __code0 function to correctly handle scenarios where 'Answer:' is not present would be a more isolated and safer change.\",\n",
      "\"answer\": \"\",\n",
      "\"suggestion\": {\n",
      "\"__code0\": \"def extract_answer(self, prompt_template, question, response):\\n    if 'Answer:' in response:\\n        answer = response.split('Answer:')[1].strip()\\n    else:\\n        answer = 'The response does not contain a clear answer.'\\n    return answer\"\n",
      "}\n",
      "}\n",
      "Question: Is the following sentence plausible? \"Marcelo got on the end of a through ball.\"\n",
      "Expected answer: yes\n",
      "Answer: MessageNode: (eval:1, dtype=<class 'str'>, data=The response does not contain a clear answer.)\n",
      "Output: MessageNode: (eval:1, dtype=<class 'str'>, data=The response does not contain a clear answer.), Feedback: The answer is wrong. We expect the output of your answer to be \"yes\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer., Variables:\n",
      "__code:1 def create_prompt(self, prompt_template, question):\n",
      "       return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question}'\n",
      "__code:0 def extract_answer(self, prompt_template, question, response):\n",
      "    if 'Answer:' in response:\n",
      "        answer = response.split('Answer:')[1].strip()\n",
      "    else:\n",
      "        answer = 'The response does not contain a clear answer.'\n",
      "    return answer\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "Prompt\n",
      " \n",
      "You're tasked to solve a coding/algorithm problem. You will see the instruction, the code, the documentation of each function used in the code, and the feedback about the execution result.\n",
      "\n",
      "Specifically, a problem will be composed of the following parts:\n",
      "- #Instruction: the instruction which describes the things you need to do or the question you should answer.\n",
      "- #Code: the code defined in the problem.\n",
      "- #Documentation: the documentation of each function used in #Code. The explanation might be incomplete and just contain high-level description. You can use the values in #Others to help infer how those functions work.\n",
      "- #Variables: the input variables that you can change.\n",
      "- #Constraints: the constraints or descriptions of the variables in #Variables.\n",
      "- #Inputs: the values of other inputs to the code, which are not changeable.\n",
      "- #Others: the intermediate values created through the code execution.\n",
      "- #Outputs: the result of the code output.\n",
      "- #Feedback: the feedback about the code's execution result.\n",
      "\n",
      "In #Variables, #Inputs, #Outputs, and #Others, the format is:\n",
      "\n",
      "<data_type> <variable_name> = <value>\n",
      "\n",
      "If <type> is (code), it means <value> is the source code of a python code, which may include docstring and definitions.\n",
      "\n",
      "Output_format: Your output should be in the following json format, satisfying the json syntax:\n",
      "\n",
      "{{\n",
      "\"reasoning\": <Your reasoning>,\n",
      "\"answer\": <Your answer>,\n",
      "\"suggestion\": {{\n",
      "    <variable_1>: <suggested_value_1>,\n",
      "    <variable_2>: <suggested_value_2>,\n",
      "}}\n",
      "}}\n",
      "\n",
      "In \"reasoning\", explain the problem: 1. what the #Instruction means 2. what the #Feedback on #Output means to #Variables considering how #Variables are used in #Code and other values in #Documentation, #Inputs, #Others. 3. Reasoning about the suggested changes in #Variables (if needed) and the expected result.\n",
      "\n",
      "If #Instruction asks for an answer, write it down in \"answer\".\n",
      "\n",
      "If you need to suggest a change in the values of #Variables, write down the suggested values in \"suggestion\". Remember you can change only the values in #Variables, not others. When <type> of a variable is (code), you should write the new definition in the format of python code without syntax errors, and you should not change the function name or the function signature.\n",
      "\n",
      "If no changes or answer are needed, just output TERMINATE.\n",
      "\n",
      "Now you see problem instance:\n",
      "\n",
      "================================\n",
      "\n",
      "#Instruction\n",
      "You need to change the <value> of the variables in #Variables to improve the output in accordance to #Feedback.\n",
      "\n",
      "#Code\n",
      "eval0 = eval(self=ModelWrapper0, prompt_template=str0, question=str0, __code=__code1)\n",
      "LLMCallable.call_llm0 = LLMCallable.call_llm(self=ModelWrapper1, user_prompt=eval0)\n",
      "eval1 = eval(self=ModelWrapper2, prompt_template=str0, question=str1, response=LLMCallable.call_llm0, __code=__code0)\n",
      "\n",
      "#Documentation\n",
      "[eval] This operator eval(__code, *args, **kwargs) evaluates the code block, where __code is the code (str) and *args and **kwargs are the arguments of the function. The output is the result of the evaluation, i.e., __code(*args, **kwargs).\n",
      "[LLMCallable.call_llm] .\n",
      "\n",
      "#Variables\n",
      "(str) str0=Is the following sentence plausible? \"Marcelo got on the end of a through ball.\"\n",
      "(code) __code1:def create_prompt(self, prompt_template, question):\n",
      "       return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question}'\n",
      "(code) __code0:def extract_answer(self, prompt_template, question, response):\n",
      "    if 'Answer:' in response:\n",
      "        answer = response.split('Answer:')[1].strip()\n",
      "    else:\n",
      "        answer = 'The response does not contain a clear answer.'\n",
      "    return answer\n",
      "\n",
      "#Constraints\n",
      "(code) __code1: The code should start with:\n",
      "def create_prompt(self, prompt_template, question):\n",
      "(code) __code0: The code should start with:\n",
      "def extract_answer(self, prompt_template, question, response):\n",
      "\n",
      "#Inputs\n",
      "(ModelWrapper) ModelWrapper2=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(str) str1=Is the following sentence plausible? \"Marcelo got on the end of a through ball.\"\n",
      "(ModelWrapper) ModelWrapper1=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(ModelWrapper) ModelWrapper0=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "\n",
      "#Others\n",
      "(str) eval0=Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\" With a focus on the likely scenarios, evaluate: Is the following sentence plausible? \"Marcelo got on the end of a through ball.\"\n",
      "(str) LLMCallable.call_llm0=To analyze the plausibility of these sentences considering known historical contexts and logical sequences, let's break them down based on the contexts implied by their activities and the specific phrasing used.\n",
      "\n",
      "1. **\"Elias Lindhorn beat the buzzer.\"**: This sentence is typical in the context of sports, particularly those played with a clock, such as basketball, where the term \"beat the buzzer\" implies making a score just before the game or quarter clock expires. There's nothing inherently implausible about this statement without additional context. The name \"Elias Lindhorn,\" while not immediately recognizable as a famous athlete from history as of my last knowledge update, does not detract from the plausibility. Sports players regularly achieve significant moments, and new talents are constantly emerging. Therefore, without specific historical or logical contradictions, this sentence is plausible.\n",
      "\n",
      "2. **\"Marcelo got on the end of a through ball.\"**: This sentence fits well within the context of soccer (football), where a \"through ball\" is a pass sent to a player in a position to score or advance the ball significantly, often bypassing defenders. The name \"Marcelo\" is commonly associated with soccer, notably with Marcelo Vieira da Silva Júnior, a Brazilian footballer known for his tenure at Real Madrid among other achievements. Even without referencing this specific Marcelo, the scenario described is a routine part of a soccer match, making the sentence highly plausible. Soccer is the most popular sport globally, and the tactics described align with common gameplay. \n",
      "\n",
      "In conclusion, both sentences are plausible within the contexts of their respective sports, assuming no extraordinary circumstances or inaccuracies in the details provided.\n",
      "\n",
      "#Outputs\n",
      "(str) eval1=The response does not contain a clear answer.\n",
      "\n",
      "#Feedback:\n",
      "The answer is wrong. We expect the output of your answer to be \"yes\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer.\n",
      "\n",
      "================================\n",
      "\n",
      "\n",
      "Your response:\n",
      "\n",
      "LLM response:\n",
      " {\n",
      "\"reasoning\": \"From the given problem, the output indicates that the function extract_answer could not find a clear and definite 'Answer:' formulated within the response provided by LLMCallable.call_llm0. However, the response from call_llm0 does analyze the plausibility of the sentences in detail, concluding that both are plausible within their contexts. The issue seems to be that the provided response does not adhere to a format expected by the extract_answer function, which expects the response to include a specific 'Answer:' preamble followed by the conclusion. Since the feedback mentions that the answer should be 'yes', the prompt creation code (__code1) could be adapted to direct the Large Language Model (LLM) to formulate responses in a way that includes a definitive 'Answer: yes' or 'Answer: no', based on the analysis of sentence plausibility. Specifically, modifying the create_prompt function to include guidelines for response formatting may help the LLM to generate an answer that fits the expected format by the extract_answer function.\",\n",
      "\"answer\": \"\",\n",
      "\"suggestion\": {\n",
      "    \"__code1\": \"def create_prompt(self, prompt_template, question):\\n    return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question} Please provide your conclusion with a definitive \\\"Answer: yes\\\" or \\\"Answer: no\\\".'\"\n",
      "}\n",
      "}\n",
      "Question: Is the following sentence plausible? \"Deshaun Watson was called for the goal tend in the Eastern Conference Finals.\"\n",
      "Expected answer: no\n",
      "Answer: MessageNode: (eval:1, dtype=<class 'str'>, data=yes)\n",
      "Output: MessageNode: (eval:1, dtype=<class 'str'>, data=yes), Feedback: The answer is wrong. We expect the output of your answer to be \"no\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer., Variables:\n",
      "__code:1 def create_prompt(self, prompt_template, question):\n",
      "    return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question} Please provide your conclusion with a definitive \"Answer: yes\" or \"Answer: no\".'\n",
      "__code:0 def extract_answer(self, prompt_template, question, response):\n",
      "    if 'Answer:' in response:\n",
      "        answer = response.split('Answer:')[1].strip()\n",
      "    else:\n",
      "        answer = 'The response does not contain a clear answer.'\n",
      "    return answer\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "str:0 Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\"\n",
      "Prompt\n",
      " \n",
      "You're tasked to solve a coding/algorithm problem. You will see the instruction, the code, the documentation of each function used in the code, and the feedback about the execution result.\n",
      "\n",
      "Specifically, a problem will be composed of the following parts:\n",
      "- #Instruction: the instruction which describes the things you need to do or the question you should answer.\n",
      "- #Code: the code defined in the problem.\n",
      "- #Documentation: the documentation of each function used in #Code. The explanation might be incomplete and just contain high-level description. You can use the values in #Others to help infer how those functions work.\n",
      "- #Variables: the input variables that you can change.\n",
      "- #Constraints: the constraints or descriptions of the variables in #Variables.\n",
      "- #Inputs: the values of other inputs to the code, which are not changeable.\n",
      "- #Others: the intermediate values created through the code execution.\n",
      "- #Outputs: the result of the code output.\n",
      "- #Feedback: the feedback about the code's execution result.\n",
      "\n",
      "In #Variables, #Inputs, #Outputs, and #Others, the format is:\n",
      "\n",
      "<data_type> <variable_name> = <value>\n",
      "\n",
      "If <type> is (code), it means <value> is the source code of a python code, which may include docstring and definitions.\n",
      "\n",
      "Output_format: Your output should be in the following json format, satisfying the json syntax:\n",
      "\n",
      "{{\n",
      "\"reasoning\": <Your reasoning>,\n",
      "\"answer\": <Your answer>,\n",
      "\"suggestion\": {{\n",
      "    <variable_1>: <suggested_value_1>,\n",
      "    <variable_2>: <suggested_value_2>,\n",
      "}}\n",
      "}}\n",
      "\n",
      "In \"reasoning\", explain the problem: 1. what the #Instruction means 2. what the #Feedback on #Output means to #Variables considering how #Variables are used in #Code and other values in #Documentation, #Inputs, #Others. 3. Reasoning about the suggested changes in #Variables (if needed) and the expected result.\n",
      "\n",
      "If #Instruction asks for an answer, write it down in \"answer\".\n",
      "\n",
      "If you need to suggest a change in the values of #Variables, write down the suggested values in \"suggestion\". Remember you can change only the values in #Variables, not others. When <type> of a variable is (code), you should write the new definition in the format of python code without syntax errors, and you should not change the function name or the function signature.\n",
      "\n",
      "If no changes or answer are needed, just output TERMINATE.\n",
      "\n",
      "Now you see problem instance:\n",
      "\n",
      "================================\n",
      "\n",
      "#Instruction\n",
      "You need to change the <value> of the variables in #Variables to improve the output in accordance to #Feedback.\n",
      "\n",
      "#Code\n",
      "eval0 = eval(self=ModelWrapper0, prompt_template=str0, question=str0, __code=__code1)\n",
      "LLMCallable.call_llm0 = LLMCallable.call_llm(self=ModelWrapper1, user_prompt=eval0)\n",
      "eval1 = eval(self=ModelWrapper2, prompt_template=str0, question=str1, response=LLMCallable.call_llm0, __code=__code0)\n",
      "\n",
      "#Documentation\n",
      "[eval] This operator eval(__code, *args, **kwargs) evaluates the code block, where __code is the code (str) and *args and **kwargs are the arguments of the function. The output is the result of the evaluation, i.e., __code(*args, **kwargs).\n",
      "[LLMCallable.call_llm] .\n",
      "\n",
      "#Variables\n",
      "(str) str0=Is the following sentence plausible? \"Deshaun Watson was called for the goal tend in the Eastern Conference Finals.\"\n",
      "(code) __code1:def create_prompt(self, prompt_template, question):\n",
      "    return f'{prompt_template} With a focus on the likely scenarios, evaluate: {question} Please provide your conclusion with a definitive \"Answer: yes\" or \"Answer: no\".'\n",
      "(code) __code0:def extract_answer(self, prompt_template, question, response):\n",
      "    if 'Answer:' in response:\n",
      "        answer = response.split('Answer:')[1].strip()\n",
      "    else:\n",
      "        answer = 'The response does not contain a clear answer.'\n",
      "    return answer\n",
      "\n",
      "#Constraints\n",
      "(code) __code1: The code should start with:\n",
      "def create_prompt(self, prompt_template, question):\n",
      "(code) __code0: The code should start with:\n",
      "def extract_answer(self, prompt_template, question, response):\n",
      "\n",
      "#Inputs\n",
      "(ModelWrapper) ModelWrapper2=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(str) str1=Is the following sentence plausible? \"Deshaun Watson was called for the goal tend in the Eastern Conference Finals.\"\n",
      "(ModelWrapper) ModelWrapper1=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "(ModelWrapper) ModelWrapper0=<opto.trace.modules.model.<locals>.ModelWrapper object at 0x000002706C4DABB0>\n",
      "\n",
      "#Others\n",
      "(str) eval0=Considering known historical contexts and logical sequences, is the following sentence plausible? \"Elias Lindhorn beat the buzzer.\" With a focus on the likely scenarios, evaluate: Is the following sentence plausible? \"Deshaun Watson was called for the goal tend in the Eastern Conference Finals.\" Please provide your conclusion with a definitive \"Answer: yes\" or \"Answer: no\".\n",
      "(str) LLMCallable.call_llm0=Answer: yes\n",
      "\n",
      "#Outputs\n",
      "(str) eval1=yes\n",
      "\n",
      "#Feedback:\n",
      "The answer is wrong. We expect the output of your answer to be \"no\". Please modify the prompt and relevant parts of the program to help LLM produce the right answer.\n",
      "\n",
      "================================\n",
      "\n",
      "\n",
      "Your response:\n",
      "\n",
      "LLM response:\n",
      " {\n",
      "\"reasoning\": \"The instruction indicates that the output of our program is incorrect; specifically, the final evaluation ('eval1') of whether Deshaun Watson being called for the goal tend in the Eastern Conference Finals is plausible produced an incorrect 'yes' answer when the expected answer is 'no'. The problem lies within the formulation of the prompt passed to the language model. The prompt is generated by the '__code1' function and is used to ask the language model for an evaluation. Given the feedback, it's clear that the generated prompt does not correctly guide the language model to consider the implausibility of Deshaun Watson being called for a 'goal tend' which is a term from basketball, not applicable in contexts where Watson, a football player, would be involved. Therefore, the corrective action involves refining the '__code1' function to ensure that the language model not only focuses on 'likely scenarios' but also has enough context to understand and analyze the scenario's plausibility accurately, especially regarding sports context.\",\n",
      "\"answer\": \"\",\n",
      "\"suggestion\": {\n",
      "  \"__code1\": \"def create_prompt(self, prompt_template, question):\\n    return f'{prompt_template} Considering the context of the scenario, evaluate the plausibility based on the involved sports figures and terms: {question} Your response should conclude with a definitive \\\"Answer: yes\\\" or \\\"Answer: no\\\".'\"\n",
      "}\n",
      "}\n",
      "Question: Is the following sentence plausible? \"Mookie Betts skated behind the net.\"\n",
      "Expected answer: no\n",
      "Answer: MessageNode: (eval:1, dtype=<class 'str'>, data=no)\n",
      "\n",
      "Testing on new examples:\n",
      "Question: Is the following sentence plausible? \"John Tavares earned a trip to the penalty box in the Stanley Cup.\"\n",
      "Expected answer: yes\n",
      "Answer: yes\n"
     ]
    }
   ],
   "source": [
    "task = \"sports_understanding\"\n",
    "train = load_dataset(\"maveriq/bigbenchhard\", task)[\"train\"]\n",
    "examples = [{\"question\": r[\"input\"], \"answer\": r[\"target\"]} for r in train]\n",
    "\n",
    "dp = Predict()\n",
    "optimizer = OptoPrime(dp.parameters(),\n",
    "                                    config_list=autogen.config_list_from_json(\"OAI_CONFIG_LIST\"))\n",
    "\n",
    "print(\"Training on a few examples:\")\n",
    "learn_predict(dp, optimizer, examples[:5])\n",
    "    \n",
    "print(\"\\nTesting on new examples:\")\n",
    "for example in examples[5:6]:\n",
    "    try:\n",
    "        response = dp.forward(example[\"question\"])\n",
    "        print(\"Question:\", example[\"question\"])\n",
    "        print(\"Expected answer:\", example[\"answer\"])\n",
    "        print(\"Answer:\", response.data)\n",
    "    except ExecutionError as e:\n",
    "        print(\"Question:\", example[\"question\"])\n",
    "        print(\"Expected answer:\", example[\"answer\"])\n",
    "        print(\"Error:\", e.exception_node.data)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Now, you can run each cell in this notebook step by step to walk through the process of setting up and optimizing prompts for the trading game. Happy optimizing!"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.19"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
